"""RFC 2217 (Telnet Com Port Control) implementation."""

from __future__ import annotations

from abc import ABC, abstractmethod
import dataclasses
from enum import IntEnum, IntFlag
import logging
from typing import ClassVar

from typing_extensions import Self

LOGGER = logging.getLogger(__name__)


class TelnetOption(IntEnum):
    """Telnet option codes."""

    BINARY = 0
    ECHO = 1
    SUPPRESS_GO_AHEAD = 3
    COM_PORT_OPTION = 44

    @classmethod
    def _missing_(cls, value: object) -> Self:
        """Allow unknown option codes to pass through as int values."""
        assert isinstance(value, int)
        obj = int.__new__(cls, value)
        obj._name_ = f"UNKNOWN_{value}"
        obj._value_ = value
        return obj


class TelnetCmdId(IntEnum):
    """Telnet command bytes (follow IAC)."""

    SE = 240
    NOP = 241
    DM = 242
    BRK = 243
    IP = 244
    AO = 245
    AYT = 246
    EC = 247
    EL = 248
    GA = 249
    SB = 250
    WILL = 251
    WONT = 252  # codespell:ignore wont
    DO = 253
    DONT = 254
    IAC = 255


class Rfc2217CmdId(IntEnum):
    """RFC 2217 subnegotiation command codes (client to server)."""

    SIGNATURE = 0
    SET_BAUDRATE = 1
    SET_DATASIZE = 2
    SET_PARITY = 3
    SET_STOPSIZE = 4
    SET_CONTROL = 5
    NOTIFY_LINESTATE = 6
    NOTIFY_MODEMSTATE = 7
    FLOWCONTROL_SUSPEND = 8
    FLOWCONTROL_RESUME = 9
    SET_LINESTATE_MASK = 10
    SET_MODEMSTATE_MASK = 11
    PURGE_DATA = 12


class ControlCmdId(IntEnum):
    """Values for SET_CONTROL command."""

    # Outbound/Both flow control
    REQ_FLOW_CONTROL = 0
    USE_NO_FLOW_CONTROL = 1
    USE_XON_XOFF = 2
    USE_HARDWARE = 3

    # Break
    REQ_BREAK_STATE = 4
    SET_BREAK_ON = 5
    SET_BREAK_OFF = 6

    # DTR
    REQ_DTR_STATE = 7
    SET_DTR_ON = 8
    SET_DTR_OFF = 9

    # RTS
    REQ_RTS_STATE = 10
    SET_RTS_ON = 11
    SET_RTS_OFF = 12

    # Inbound flow control
    REQ_FLOW_CONTROL_IN = 13
    USE_NO_FLOW_CONTROL_IN = 14
    USE_XON_XOFF_IN = 15
    USE_HARDWARE_IN = 16
    USE_DCD_FLOW_CONTROL = 17
    USE_DTR_FLOW_CONTROL_IN = 18
    USE_DSR_FLOW_CONTROL = 19


class Rfc2217Parity(IntEnum):
    """RFC 2217 parity values for SET_PARITY."""

    REQUEST = 0
    NONE = 1
    ODD = 2
    EVEN = 3
    MARK = 4
    SPACE = 5


class Rfc2217StopSize(IntEnum):
    """RFC 2217 stop bit values for SET_STOPSIZE."""

    REQUEST = 0
    ONE = 1
    TWO = 2
    ONE_POINT_FIVE = 3


class LineStateFlag(IntFlag):
    """Bitmask for NOTIFY_LINESTATE and SET_LINESTATE_MASK."""

    DATA_READY = 1
    OVERRUN_ERROR = 2
    PARITY_ERROR = 4
    FRAMING_ERROR = 8
    BREAK_DETECT = 16
    TRANSFER_HOLDING_REG_EMPTY = 32
    TRANSFER_SHIFT_REG_EMPTY = 64
    TIMEOUT_ERROR = 128


class ModemStateFlag(IntFlag):
    """Bitmask for NOTIFY_MODEMSTATE and SET_MODEMSTATE_MASK."""

    DELTA_CTS = 1
    DELTA_DSR = 2
    TRAILING_EDGE_RI = 4
    DELTA_RLSD = 8
    CTS = 16
    DSR = 32
    RI = 64
    RLSD = 128  # Carrier Detect


class PurgeDataValue(IntEnum):
    """Values for PURGE_DATA command."""

    RECEIVE = 1
    TRANSMIT = 2
    BOTH = 3


class BaseCommand(ABC):
    """Base class for commands."""

    CMD_ID: ClassVar[int]

    @abstractmethod
    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        raise NotImplementedError


# ---------------------------------------------------------------------------
# Telnet-level commands (IAC <cmd> <option>)
# ---------------------------------------------------------------------------


class TelnetCommand(BaseCommand):
    """Base class for telnet commands."""

    CMD_ID: ClassVar[TelnetCmdId]


@dataclasses.dataclass(frozen=True, kw_only=True)
class TelnetOptionCommand(TelnetCommand, ABC):
    """Base class for telnet option negotiation commands."""

    option: TelnetOption

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return bytes([self.option])

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(option=TelnetOption(payload[0]))


@dataclasses.dataclass(frozen=True, kw_only=True)
class WillCmd(TelnetOptionCommand):
    """IAC WILL <option> — offer to perform an option."""

    CMD_ID: ClassVar[TelnetCmdId] = TelnetCmdId.WILL


@dataclasses.dataclass(frozen=True, kw_only=True)
class WontCmd(TelnetOptionCommand):
    """IAC WONT <option> — refuse to perform an option."""  # codespell:ignore wont

    CMD_ID: ClassVar[TelnetCmdId] = TelnetCmdId.WONT  # codespell:ignore wont


@dataclasses.dataclass(frozen=True, kw_only=True)
class DoCmd(TelnetOptionCommand):
    """IAC DO <option> — request the other side perform an option."""

    CMD_ID: ClassVar[TelnetCmdId] = TelnetCmdId.DO


@dataclasses.dataclass(frozen=True, kw_only=True)
class DontCmd(TelnetOptionCommand):
    """IAC DONT <option> — demand the other side stop performing an option."""

    CMD_ID: ClassVar[TelnetCmdId] = TelnetCmdId.DONT


# ---------------------------------------------------------------------------
# RFC 2217 subnegotiation commands (IAC SB 44 <cmd> <payload> IAC SE)
# ---------------------------------------------------------------------------


class Rfc2217Command(BaseCommand):
    """Base class for RFC 2217 subnegotiation commands."""

    CMD_ID: ClassVar[Rfc2217CmdId]


@dataclasses.dataclass(frozen=True, kw_only=True)
class SignatureCmd(Rfc2217Command):
    """Exchange signature/identity strings. Empty = request."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.SIGNATURE
    signature: bytes = b""

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return self.signature

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(signature=payload)


@dataclasses.dataclass(frozen=True, kw_only=True)
class SetBaudrateCmd(Rfc2217Command):
    """Set baud rate. 0 = query current value."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.SET_BAUDRATE
    baudrate: int

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return self.baudrate.to_bytes(4, "big")

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(baudrate=int.from_bytes(payload, "big"))


@dataclasses.dataclass(frozen=True, kw_only=True)
class SetDatasizeCmd(Rfc2217Command):
    """Set data bit size. 0 = query, 5-8 = actual size."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.SET_DATASIZE
    size: int

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return self.size.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(size=payload[0])


@dataclasses.dataclass(frozen=True, kw_only=True)
class SetParityCmd(Rfc2217Command):
    """Set parity."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.SET_PARITY
    parity: Rfc2217Parity

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return self.parity.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(parity=Rfc2217Parity(payload[0]))


@dataclasses.dataclass(frozen=True, kw_only=True)
class SetStopsizeCmd(Rfc2217Command):
    """Set stop bits."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.SET_STOPSIZE
    size: Rfc2217StopSize

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return self.size.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(size=Rfc2217StopSize(payload[0]))


@dataclasses.dataclass(frozen=True, kw_only=True)
class SetControlCmd(Rfc2217Command):
    """Set flow control, break, DTR, or RTS."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.SET_CONTROL
    control: ControlCmdId

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return self.control.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(control=ControlCmdId(payload[0]))


@dataclasses.dataclass(frozen=True, kw_only=True)
class NotifyLinestateCmd(Rfc2217Command):
    """Server notification of UART line state change."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.NOTIFY_LINESTATE
    linestate: LineStateFlag

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return self.linestate.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(linestate=LineStateFlag(payload[0]))


@dataclasses.dataclass(frozen=True, kw_only=True)
class NotifyModemstateCmd(Rfc2217Command):
    """Server notification of modem state change."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.NOTIFY_MODEMSTATE
    modemstate: ModemStateFlag

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return self.modemstate.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(modemstate=ModemStateFlag(payload[0]))


@dataclasses.dataclass(frozen=True, kw_only=True)
class FlowcontrolSuspendCmd(Rfc2217Command):
    """Request the receiver suspend transmission."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.FLOWCONTROL_SUSPEND

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return b""

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls()


@dataclasses.dataclass(frozen=True, kw_only=True)
class FlowcontrolResumeCmd(Rfc2217Command):
    """Request the receiver resume transmission."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.FLOWCONTROL_RESUME

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return b""

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls()


@dataclasses.dataclass(frozen=True, kw_only=True)
class SetLinestateMaskCmd(Rfc2217Command):
    """Set which line state changes trigger NOTIFY_LINESTATE. Default: 0."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.SET_LINESTATE_MASK
    mask: LineStateFlag

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return self.mask.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(mask=LineStateFlag(payload[0]))


@dataclasses.dataclass(frozen=True, kw_only=True)
class SetModemstateMaskCmd(Rfc2217Command):
    """Set which modem state changes trigger NOTIFY_MODEMSTATE. Default: 255."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.SET_MODEMSTATE_MASK
    mask: ModemStateFlag

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return self.mask.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(mask=ModemStateFlag(payload[0]))


@dataclasses.dataclass(frozen=True, kw_only=True)
class PurgeDataCmd(Rfc2217Command):
    """Purge access server data buffers."""

    CMD_ID: ClassVar[Rfc2217CmdId] = Rfc2217CmdId.PURGE_DATA
    what: PurgeDataValue

    def to_bytes(self) -> bytes:
        """Serialize to bytes."""
        return self.what.to_bytes(1, "big")

    @classmethod
    def from_bytes(cls, payload: bytes) -> Self:
        """Parse from bytes."""
        return cls(what=PurgeDataValue(payload[0]))


CMD_ID_TO_CLASS: dict[Rfc2217CmdId, type[Rfc2217Command]] = {
    cls.CMD_ID: cls
    for cls in (
        SignatureCmd,
        SetBaudrateCmd,
        SetDatasizeCmd,
        SetParityCmd,
        SetStopsizeCmd,
        SetControlCmd,
        NotifyLinestateCmd,
        NotifyModemstateCmd,
        FlowcontrolSuspendCmd,
        FlowcontrolResumeCmd,
        SetLinestateMaskCmd,
        SetModemstateMaskCmd,
        PurgeDataCmd,
    )
}


def iac_escape(data: bytes) -> bytes:
    """Double any 0xFF bytes so they aren't mistaken for IAC."""
    return data.replace(b"\xff", b"\xff\xff")


def encode_command(cmd: TelnetCommand | Rfc2217Command) -> bytes:
    """Encode a command to its wire representation.

    For telnet commands: IAC <cmd> <option>
    For RFC 2217 commands: IAC SB 44 <cmd_id> <payload> IAC SE
    """

    if isinstance(cmd, TelnetCommand):
        encoded = bytes([TelnetCmdId.IAC, cmd.CMD_ID]) + cmd.to_bytes()
    else:
        encoded = (
            bytes(
                [
                    TelnetCmdId.IAC,
                    TelnetCmdId.SB,
                    TelnetOption.COM_PORT_OPTION,
                    cmd.CMD_ID,
                ]
            )
            + iac_escape(cmd.to_bytes())
            + bytes([TelnetCmdId.IAC, TelnetCmdId.SE])
        )

    return encoded



"""Serial port communication utilities."""

from __future__ import annotations

from abc import abstractmethod
import asyncio
from asyncio import IncompleteReadError
import bisect
from collections import defaultdict
from collections.abc import Awaitable, Callable, Iterator
from contextlib import contextmanager
import dataclasses
from enum import Enum
import functools
import io
import os.path
from pathlib import Path
import time
from types import TracebackType
from typing import Any, Concatenate, NamedTuple, ParamSpec, TypeVar, cast
import urllib.parse
import warnings

from typing_extensions import Buffer, Self, TypedDict, Unpack


class Platform(str, Enum):
    """Built-in platform name."""

    DEVICE = "device"

    POSIX = "posix"
    EXTENDED_POSIX = "extended_posix"
    FREEBSD = "freebsd"
    LINUX = "linux"
    DARWIN = "darwin"
    WIN32 = "win32"
    SOCKET = "socket"
    RFC2217 = "rfc2217"
    ESPHOME = "esphome"


@dataclasses.dataclass(frozen=True)
class RegisteredUriHandler:
    """A URI handler registration entry."""

    scheme: str
    unique_scheme: str
    weight: int
    sync_cls: type[BaseSerial]
    async_transport_cls: type[BaseSerialTransport]
    list_serial_ports_func: Callable[..., list[SerialPortInfo]]
    async_list_serial_ports_func: Callable[..., Awaitable[list[SerialPortInfo]]]
    strip_uri_scheme: bool


class _RegistryEntry(NamedTuple):
    """Entry in `_REGISTERED_URI_HANDLERS`, ordered by (weight, insertion_time)."""

    weight: int
    insertion_time: float  # To avoid comparing `RegisteredUriHandler` objects
    handler: RegisteredUriHandler


_REGISTERED_URI_HANDLERS: defaultdict[str, list[_RegistryEntry]] = defaultdict(list)


def empty_port_list(*args: Any, **kwargs: Any) -> list[SerialPortInfo]:
    """Return an empty list of serial ports."""
    return []


async def async_empty_port_list(*args: Any, **kwargs: Any) -> list[SerialPortInfo]:
    """Return an empty list of serial ports, async."""
    return []


def register_uri_handler(
    *,
    scheme: str,
    unique_scheme: str,
    sync_cls: type[BaseSerial],
    async_transport_cls: type[BaseSerialTransport],
    list_serial_ports_func: Callable[..., list[SerialPortInfo]] = empty_port_list,
    async_list_serial_ports_func: Callable[
        ..., Awaitable[list[SerialPortInfo]]
    ] = async_empty_port_list,
    weight: int = 1,
    strip_uri_scheme: bool = False,
) -> Callable[[], None]:
    """Register a URI handler.

    Expose a new backend to ``serial_for_url`` / ``create_serial_connection`` /
    ``open_serial_connection``.

    Args:
        scheme: Shared dispatch scheme. URLs with this scheme resolve to the
            highest-weight handler registered under it.
        unique_scheme: A scheme that uniquely identifies this handler. Must end
            with ``://`` and must not collide with an existing registration.
            Use this to address the handler directly.
        sync_cls: Synchronous serial class, typically a subclass of
            :class:`BaseSerial`.
        async_transport_cls: Async transport class, typically a subclass of
            :class:`BaseSerialTransport`.
        list_serial_ports_func: Callable returning a list of :class:`SerialPortInfo`.
        async_list_serial_ports_func: Async callable returning a list of
            :class:`SerialPortInfo`.
        weight: Dispatch priority under ``scheme``. Higher wins.
        strip_uri_scheme: If ``True``, the leading ``scheme`` / ``unique_scheme``
            is removed before the URL is passed to the sync class. Set this when
            the underlying class expects a bare device path rather than a URL.

    Returns:
        A callable that unregisters the handler.

    Raises:
        ValueError: if either scheme doesn't end with ``://`` or ``unique_scheme``
            is already registered.

    """
    if not scheme.endswith("://") or not unique_scheme.endswith("://"):
        raise ValueError(f"Schemes {scheme!r} and {unique_scheme!r} must end with ://")

    if _REGISTERED_URI_HANDLERS[unique_scheme]:
        raise ValueError(
            f"URI scheme {unique_scheme!r} is not unique,"
            f" already registered to {_REGISTERED_URI_HANDLERS[unique_scheme]}"
        )

    item = _RegistryEntry(
        weight=weight,
        insertion_time=time.monotonic(),
        handler=RegisteredUriHandler(
            scheme=scheme,
            unique_scheme=unique_scheme,
            weight=weight,
            sync_cls=sync_cls,
            async_transport_cls=async_transport_cls,
            list_serial_ports_func=list_serial_ports_func,
            async_list_serial_ports_func=async_list_serial_ports_func,
            strip_uri_scheme=strip_uri_scheme,
        ),
    )
    bisect.insort_right(_REGISTERED_URI_HANDLERS[scheme], item)

    if unique_scheme != scheme:
        _REGISTERED_URI_HANDLERS[unique_scheme].append(item)

    def remove_callback() -> None:
        _REGISTERED_URI_HANDLERS[scheme].remove(item)
        if unique_scheme != scheme:
            _REGISTERED_URI_HANDLERS[unique_scheme].remove(item)

    return remove_callback


def get_uri_handler(uri: str) -> RegisteredUriHandler:
    """Look up the registered handler for the given URI."""
    parsed_uri = urllib.parse.urlparse(uri)
    scheme = (parsed_uri.scheme or "device") + "://"
    handlers = _REGISTERED_URI_HANDLERS.get(scheme)
    if not handlers:
        raise UnknownUriScheme(f"No handler registered for URI scheme {scheme!r}")
    return handlers[-1].handler


class SerialException(Exception):
    """Base serial exception."""


class UnsupportedSetting(SerialException):
    """Raised when an unsupported serial port setting is used."""


class UnknownUriScheme(SerialException):
    """Raised when a URI scheme has no registered handler."""


class StopBits(float, Enum):
    """Stop bits configuration."""

    ONE = 1
    ONE_POINT_FIVE = 1.5
    TWO = 2


class Parity(str, Enum):
    """Parity configuration."""

    NONE = "N"
    ODD = "O"
    EVEN = "E"
    MARK = "M"
    SPACE = "S"


class ConnectKwargs(  # type: ignore[call-arg]  # PEP 728 not in mypy yet
    TypedDict, total=False, extra_items=Any
):
    """Kwargs forwarded to BaseSerialTransport.connect / _connect."""

    baudrate: int
    parity: Parity
    stopbits: StopBits
    xonxoff: bool
    rtscts: bool
    exclusive: bool
    byte_size: int


class PinState(Enum):
    """Pin state."""

    UNDEFINED = None
    LOW = 0
    HIGH = 1

    @classmethod
    def convert(cls, value: PinState | bool | None) -> PinState:
        """Create PinState from boolean."""
        if isinstance(value, cls):
            return value

        if value is None:
            return cls.UNDEFINED

        return cls.HIGH if value else cls.LOW

    def to_bool(self) -> bool | None:
        """Convert PinState to boolean."""
        if self is PinState.UNDEFINED:
            return None

        return self is PinState.HIGH


@dataclasses.dataclass(frozen=True)
class ModemPins:
    """Modem control bits."""

    le: PinState = PinState.UNDEFINED
    dtr: PinState = PinState.UNDEFINED
    rts: PinState = PinState.UNDEFINED
    st: PinState = PinState.UNDEFINED
    sr: PinState = PinState.UNDEFINED
    cts: PinState = PinState.UNDEFINED
    car: PinState = PinState.UNDEFINED
    rng: PinState = PinState.UNDEFINED
    dsr: PinState = PinState.UNDEFINED

    def __repr__(self) -> str:
        """Return string representation of modem pins."""

        bits = []

        for bit in ("le", "dtr", "rts", "st", "sr", "cts", "car", "rng", "dsr"):
            value = getattr(self, bit)

            if value is PinState.UNDEFINED:
                continue
            elif value is PinState.HIGH:
                bits.append(bit)
            else:
                bits.append(f"!{bit}")

        return f"{self.__class__.__name__}[{' '.join(bits)}]"


@contextmanager
def measure_time() -> Iterator[Callable[[], float]]:
    """Measure elapsed time in a context."""
    start: float = time.monotonic()
    end: float | None = None

    def get_result() -> float:
        if end is None:
            raise RuntimeError("Context has not exited yet")

        return end - start

    try:
        yield get_result
    finally:
        end = time.monotonic()


_P = ParamSpec("_P")
_R = TypeVar("_R")


def maybe_wrap_exceptions(
    func: Callable[Concatenate[BaseSerial, _P], _R],
) -> Callable[Concatenate[BaseSerial, _P], _R]:
    """Re-raise all exceptions as `SerialException` when the flag is set."""

    @functools.wraps(func)
    def replacement(self: BaseSerial, /, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        try:
            return func(self, *args, **kwargs)
        except Exception as exc:
            if self._wrap_exceptions:
                raise SerialException(str(exc)) from exc

            raise

    return replacement


class BaseSerial(io.RawIOBase):
    """Base class for serial port communication."""

    def __init__(
        self,
        path: str | Path | None = None,
        baudrate: int = 9600,
        *,
        parity: Parity | str | None = Parity.NONE,
        stopbits: StopBits | int | float = StopBits.ONE,
        xonxoff: bool = False,
        rtscts: bool = False,
        dsrdtr: bool = False,
        byte_size: int = 8,
        read_timeout: float | None = None,
        write_timeout: float | None = None,
        rtsdtr_on_open: PinState = PinState.HIGH,
        rtsdtr_on_close: PinState = PinState.LOW,
        exclusive: bool = True,
        # pyserial compatibility kwargs
        port: str | None = None,
        timeout: float | None = None,
        bytesize: int | None = None,
        do_not_open: bool | None = None,
        writeTimeout: float | None = None,
        inter_byte_timeout: int | None = None,
        # Internal pyserial compatibility signal
        _wrap_exceptions: bool = False,
    ) -> None:
        """Initialize serial port configuration."""
        super().__init__()

        if not isinstance(stopbits, StopBits):
            stopbits = StopBits(stopbits)

        if parity is None:
            parity = Parity.NONE
        elif not isinstance(parity, Parity):
            parity = Parity(parity)

        self._path = path
        self._baudrate = baudrate
        self._stopbits = stopbits
        self._xonxoff = xonxoff
        self._rtscts = rtscts
        self._dsrdtr = dsrdtr
        self._parity = parity
        self._byte_size = byte_size
        self._exclusive = exclusive
        self._read_timeout = read_timeout
        self._write_timeout = write_timeout

        self._rtsdtr_on_open = rtsdtr_on_open
        self._rtsdtr_on_close = rtsdtr_on_close

        self._auto_close = False

        # Compatibility kwargs
        if port is not None:
            self._path = port

        if timeout is not None:
            self._read_timeout = timeout

        if bytesize is not None:
            self._byte_size = bytesize

        if writeTimeout is not None:
            self._write_timeout = writeTimeout

        if do_not_open is False:
            raise RuntimeError("do_not_open=False is not supported")

        self._wrap_exceptions = _wrap_exceptions

        # Enter a "broken" state so that an error condition can persist
        self._broken: Exception | None = None

    def _mark_broken(self, exc: Exception) -> None:
        if self._broken is None:
            self._broken = exc

    def _check_broken(self) -> None:
        if self._broken is not None:
            raise self._broken

    @classmethod
    def from_url(cls, url: str, *args: Any, **kwargs: Any) -> BaseSerial:
        """Create the appropriate serial port subclass for the given URL."""
        handler = get_uri_handler(url)
        if handler.strip_uri_scheme:
            url = url.removeprefix(handler.scheme).removeprefix(handler.unique_scheme)
        return handler.sync_cls(url, *args, **kwargs)

    @maybe_wrap_exceptions
    def open(self) -> None:
        """Open the serial port."""
        self._broken = None
        try:
            self._open()
            self._configure_port()
        except BaseException:
            self.close()
            raise

    @maybe_wrap_exceptions
    def configure_port(self) -> None:
        """Configure the serial port settings."""
        self._configure_port()

    @abstractmethod
    def _open(self) -> None:
        """Open the serial port (platform-specific)."""
        raise NotImplementedError

    @abstractmethod
    def _configure_port(self) -> None:
        """Configure the serial port settings (platform-specific)."""
        raise NotImplementedError

    @maybe_wrap_exceptions
    def close(self) -> None:
        """Close the serial port."""
        self._close()

    @abstractmethod
    def _close(self) -> None:
        """Close the serial port, internal."""
        raise NotImplementedError

    @property
    def read_timeout(self) -> float | None:
        """Get the read timeout in seconds."""
        return self._read_timeout

    @property
    def write_timeout(self) -> float | None:
        """Get the write timeout in seconds."""
        return self._write_timeout

    @maybe_wrap_exceptions
    def get_modem_pins(self) -> ModemPins:
        """Get modem control bits."""
        self._check_broken()
        return self._get_modem_pins()

    @maybe_wrap_exceptions
    def set_modem_pins(
        self,
        modem_pins: ModemPins | None = None,
        *,
        le: PinState | bool | None = PinState.UNDEFINED,
        dtr: PinState | bool | None = PinState.UNDEFINED,
        rts: PinState | bool | None = PinState.UNDEFINED,
        st: PinState | bool | None = PinState.UNDEFINED,
        sr: PinState | bool | None = PinState.UNDEFINED,
        cts: PinState | bool | None = PinState.UNDEFINED,
        car: PinState | bool | None = PinState.UNDEFINED,
        rng: PinState | bool | None = PinState.UNDEFINED,
        dsr: PinState | bool | None = PinState.UNDEFINED,
    ) -> None:
        """Set modem control bits."""
        self._check_broken()

        if modem_pins is None:
            modem_pins = ModemPins(
                le=PinState.convert(le),
                dtr=PinState.convert(dtr),
                rts=PinState.convert(rts),
                st=PinState.convert(st),
                sr=PinState.convert(sr),
                cts=PinState.convert(cts),
                car=PinState.convert(car),
                rng=PinState.convert(rng),
                dsr=PinState.convert(dsr),
            )

        return self._set_modem_pins(modem_pins)

    @abstractmethod
    def _get_modem_pins(self) -> ModemPins:
        """Get modem control bits, internal."""
        raise NotImplementedError

    @abstractmethod
    def _set_modem_pins(self, modem_pins: ModemPins) -> None:
        """Set modem control bits, internal."""
        raise NotImplementedError

    @maybe_wrap_exceptions
    def readinto(self, b: Buffer, *, timeout: float | None = None) -> int:
        """Read bytes from serial port into buffer."""
        self._check_broken()
        timeout = self._read_timeout if timeout is None else timeout
        return self._readinto(b, timeout=timeout)

    @abstractmethod
    def _readinto(self, b: Buffer, *, timeout: float | None) -> int:
        """Read bytes from serial port into buffer, internal."""
        raise NotImplementedError

    @maybe_wrap_exceptions
    def write(self, data: Buffer, *, timeout: float | None = None) -> int:
        """Write bytes to serial port."""
        self._check_broken()
        timeout = self._write_timeout if timeout is None else timeout
        return self._write(data, timeout=timeout)

    @abstractmethod
    def _write(self, data: Buffer, *, timeout: float | None) -> int:
        """Write bytes to serial port, internal."""
        raise NotImplementedError

    def flush(self) -> None:
        """Flush write buffers."""
        self._check_broken()
        self._flush()

    @abstractmethod
    def _flush(self) -> None:
        """Flush write buffers, internal."""
        raise NotImplementedError

    @property
    def path(self) -> str | Path | None:
        """Get the serial port path."""
        return self._path

    @property
    def baudrate(self) -> int:
        """Get the baud rate."""
        return self._baudrate

    @baudrate.setter
    def baudrate(self, value: int) -> None:
        """Set baud rate (deprecated)."""
        self._baudrate = value
        self._configure_port()

    @property
    def parity(self) -> Parity:
        """Get the parity."""
        return self._parity

    @property
    def byte_size(self) -> int:
        """Get the byte size."""
        return self._byte_size

    @property
    def stopbits(self) -> StopBits:
        """Get the number of stop bits."""
        return self._stopbits

    @property
    def rtsdtr_on_open(self) -> PinState:
        """Get the RTS/DTR pin state (on open) setting."""
        return self._rtsdtr_on_open

    @property
    def rtsdtr_on_close(self) -> PinState:
        """Get the RTS/DTR pin state (on close) setting."""
        return self._rtsdtr_on_close

    @property
    def exclusive(self) -> bool:
        """Get the exclusive setting."""
        return self._exclusive

    def readexactly(self, n: int, *, timeout: float | None = None) -> bytes:
        """Read exactly n bytes."""
        buffer = bytearray(n)
        view = memoryview(buffer)
        remaining = n
        timeout = self.read_timeout if timeout is None else timeout

        while remaining > 0:
            with measure_time() as get_elapsed:
                read = self.readinto(view, timeout=timeout)

            if timeout is not None:
                timeout -= get_elapsed()

            view = view[read:]
            remaining -= read

            if read == 0:
                # `IncompleteReadError` is a subclass of `EOFError`
                raise IncompleteReadError(
                    expected=n, partial=bytes(buffer[: n - remaining])
                )

        return bytes(buffer)

    def read_until(
        self,
        expected: bytes = b"\n",
        size: int | None = None,
        *,
        timeout: float | None = None,
    ) -> bytes:
        """Read until the expected sequence is found."""
        buffer = bytearray()
        expected_len = len(expected)
        timeout = self.read_timeout if timeout is None else timeout

        while True:
            with measure_time() as get_elapsed:
                byte = self.readexactly(1, timeout=timeout)

            if timeout is not None:
                timeout -= get_elapsed()

            if not byte:
                break

            buffer += byte

            if buffer[-expected_len:] == expected:
                break

            if size is not None and len(buffer) >= size:
                break

        return bytes(buffer)

    def __enter__(self) -> Self:
        """Enter context manager."""
        self.open()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit context manager."""
        self.close()

    def __del__(self) -> None:
        """Cleanup on deletion."""
        if getattr(self, "_auto_close", False):
            self.close()

    @abstractmethod
    def num_unread_bytes(self) -> int:
        """Return the number of bytes waiting to be read."""
        raise NotImplementedError

    @abstractmethod
    def num_unwritten_bytes(self) -> int:
        """Return the number of bytes waiting to be written."""
        raise NotImplementedError

    @maybe_wrap_exceptions
    def reset_read_buffer(self) -> None:
        """Reset the read buffer."""
        self._reset_read_buffer()

    @abstractmethod
    def _reset_read_buffer(self) -> None:
        """Reset the read buffer, internal."""
        raise NotImplementedError

    @maybe_wrap_exceptions
    def reset_write_buffer(self) -> None:
        """Reset the write buffer."""
        self._reset_write_buffer()

    @abstractmethod
    def _reset_write_buffer(self) -> None:
        """Reset the write buffer, internal."""
        raise NotImplementedError

    @property
    @abstractmethod
    def is_open(self) -> bool:
        """Return whether the serial port is open."""
        raise NotImplementedError

    # Deprecated aliases
    @property
    def port(self) -> str | None:
        """Deprecated alias for `path`.

        Warning:
            Deprecated, use `path` instead.

        """
        return str(self.path) if self.path is not None else None

    @property
    def portstr(self) -> str | None:
        """Deprecated alias for `path`.

        Warning:
            Deprecated, use `path` instead.

        """
        return str(self.path) if self.path is not None else None

    @property
    def timeout(self) -> float | None:
        """Deprecated alias for `read_timeout`.

        Warning:
            Deprecated, use `read_timeout` instead.

        """
        return self.read_timeout

    @timeout.setter
    def timeout(self, value: float) -> None:
        self._read_timeout = value

    @property
    def bytesize(self) -> int:
        """Deprecated alias for `byte_size`.

        Warning:
            Deprecated, use `byte_size` instead.

        """
        return self.byte_size

    @property
    def data_bits(self) -> int:
        """Deprecated alias for `byte_size`.

        Warning:
            Deprecated, use `byte_size` instead.

        """
        return self.byte_size

    @data_bits.setter
    def data_bits(self, value: int) -> None:
        """Set the byte size (deprecated)."""
        self._byte_size = value

    @property
    def stop_bits(self) -> int | float:
        """Deprecated alias for `stopbits`.

        Warning:
            Deprecated, use `stopbits` instead.

        """
        return cast(int | float, self._stopbits.value)

    @stop_bits.setter
    def stop_bits(self, value: int | float) -> None:
        """Set the number of stop bits (deprecated)."""
        self._stopbits = StopBits(value)

    @property
    def writeTimeout(self) -> float | None:
        """Deprecated alias for `write_timeout`.

        Warning:
            Deprecated, use `write_timeout` instead.

        """
        return self.write_timeout

    def reset_input_buffer(self) -> None:
        """Reset the read buffer.

        Warning:
            Deprecated, use `reset_read_buffer` instead.

        """
        self.reset_read_buffer()

    def reset_output_buffer(self) -> None:
        """Reset the write buffer.

        Warning:
            Deprecated, use `reset_write_buffer` instead.

        """
        self.reset_write_buffer()

    def flushInput(self) -> None:
        """Reset the read buffer.

        Warning:
            Deprecated, use `reset_read_buffer` instead.

        """
        self.reset_read_buffer()

    def flushOutput(self) -> None:
        """Reset the write buffer.

        Warning:
            Deprecated, use `reset_write_buffer` instead.

        """
        self.reset_write_buffer()

    @property
    def in_waiting(self) -> int:
        """Deprecated alias for `num_unread_bytes`.

        Warning:
            Deprecated, use `num_unread_bytes` instead.

        """
        return self.num_unread_bytes()

    @property
    def out_waiting(self) -> int:
        """Deprecated alias for `num_unwritten_bytes`.

        Warning:
            Deprecated, use `num_unwritten_bytes` instead.

        """
        return self.num_unwritten_bytes()

    @property
    def inWaiting(self) -> int:
        """Deprecated alias for `num_unread_bytes`.

        Warning:
            Deprecated, use `num_unread_bytes` instead.

        """
        return self.in_waiting

    def isOpen(self) -> bool:
        """Return whether the serial port is open.

        Warning:
            Deprecated, use `is_open` instead.

        """
        return self.is_open

    @property
    def dtr(self) -> bool | None:
        """Get DTR modem bit."""
        return self.get_modem_pins().dtr.to_bool()

    @dtr.setter
    def dtr(self, value: bool) -> None:
        """Set DTR modem bit."""
        self.set_modem_pins(dtr=bool(value))

    @property
    def rts(self) -> bool | None:
        """Get RTS modem bit."""
        return self.get_modem_pins().rts.to_bool()

    @rts.setter
    def rts(self, value: bool) -> None:
        """Set RTS modem bit."""
        self.set_modem_pins(rts=bool(value))

    @property
    def cts(self) -> bool | None:
        """Get CTS modem bit."""
        return self.get_modem_pins().cts.to_bool()

    @cts.setter
    def cts(self, value: bool) -> None:
        """Set CTS modem bit."""
        self.set_modem_pins(cts=bool(value))


class BaseSerialTransport(asyncio.Transport):
    """Base class for serial port asyncio transport."""

    transport_name = "serial"

    def __init__(
        self, loop: asyncio.AbstractEventLoop, protocol: asyncio.Protocol
    ) -> None:
        """Initialize serial transport."""
        super().__init__()
        self._loop = loop
        self._protocol = protocol
        self._extra: dict[str, Any] = {}

        self._serial: BaseSerial | None = None
        self._closing: bool = False
        self._closed_waiter: asyncio.Future[None] = loop.create_future()

    def _mark_broken(self, exc: Exception) -> None:
        if self._serial is not None:
            self._serial._mark_broken(exc)

    def _check_broken(self) -> None:
        if self._serial is not None:
            self._serial._check_broken()

    def is_closing(self) -> bool:
        """Return whether the transport is closing."""
        return self._closing

    def _resolve_closed_waiter(self) -> None:
        if not self._closed_waiter.done():
            self._closed_waiter.set_result(None)

    def _call_protocol_connection_lost(self, exc: Exception | None) -> None:
        try:
            self._protocol.connection_lost(exc)
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException as protocol_exc:
            self._loop.call_exception_handler(
                {
                    "message": "protocol.connection_lost() failed",
                    "exception": protocol_exc,
                    "transport": self,
                    "protocol": self._protocol,
                }
            )
        finally:
            self._resolve_closed_waiter()

    def get_protocol(self) -> asyncio.Protocol:
        """Get the protocol used by this transport."""
        return self._protocol

    def set_protocol(self, protocol: asyncio.Protocol) -> None:  # type: ignore[override]
        """Set the protocol to use with this transport."""
        self._protocol = protocol

    @property
    def serial(self) -> BaseSerial:
        """Get the serial port instance."""
        assert self._serial is not None
        return self._serial

    @property
    def baudrate(self) -> int:
        """Get the baud rate."""
        assert self._serial is not None
        return self._serial.baudrate

    @property
    def parity(self) -> Parity:
        """Get the parity."""
        assert self._serial is not None
        return self._serial.parity

    @property
    def stopbits(self) -> StopBits:
        """Get the number of stop bits."""
        assert self._serial is not None
        return self._serial.stopbits

    @property
    def byte_size(self) -> int:
        """Get the byte size."""
        assert self._serial is not None
        return self._serial.byte_size

    @property
    def exclusive(self) -> bool:
        """Get the exclusive setting."""
        assert self._serial is not None
        return self._serial.exclusive

    @abstractmethod
    async def _connect(
        self, *, path: str | None, **kwargs: Unpack[ConnectKwargs]
    ) -> None:
        """Connect to serial port."""
        raise NotImplementedError

    async def connect(
        self,
        *,
        path: str | None,
        **kwargs: Unpack[ConnectKwargs],
    ) -> None:
        """Connect to serial port."""
        if path is not None:
            handler = get_uri_handler(path)
            if handler.strip_uri_scheme:
                path = path.removeprefix(handler.scheme).removeprefix(
                    handler.unique_scheme
                )

        try:
            await self._connect(path=path, **kwargs)
        except BaseException:
            # Intentionally catch cancellation too: callers should only observe
            # connect failure/cancel after transport resources are released.
            self.close()
            await self.wait_closed()

            raise

    async def get_modem_pins(self) -> ModemPins:
        """Get modem control bits."""
        self._check_broken()
        return await self._get_modem_pins()

    @abstractmethod
    async def _get_modem_pins(self) -> ModemPins:
        """Get modem control bits, internal."""
        raise NotImplementedError

    @abstractmethod
    async def _set_modem_pins(self, modem_pins: ModemPins) -> None:
        """Set modem control bits, internal."""
        raise NotImplementedError

    async def set_modem_pins(
        self,
        modem_pins: ModemPins | None = None,
        *,
        le: PinState | bool | None = PinState.UNDEFINED,
        dtr: PinState | bool | None = PinState.UNDEFINED,
        rts: PinState | bool | None = PinState.UNDEFINED,
        st: PinState | bool | None = PinState.UNDEFINED,
        sr: PinState | bool | None = PinState.UNDEFINED,
        cts: PinState | bool | None = PinState.UNDEFINED,
        car: PinState | bool | None = PinState.UNDEFINED,
        rng: PinState | bool | None = PinState.UNDEFINED,
        dsr: PinState | bool | None = PinState.UNDEFINED,
    ) -> None:
        """Set modem control bits."""
        self._check_broken()

        if modem_pins is None:
            modem_pins = ModemPins(
                le=PinState.convert(le),
                dtr=PinState.convert(dtr),
                rts=PinState.convert(rts),
                st=PinState.convert(st),
                sr=PinState.convert(sr),
                cts=PinState.convert(cts),
                car=PinState.convert(car),
                rng=PinState.convert(rng),
                dsr=PinState.convert(dsr),
            )

        return await self._set_modem_pins(modem_pins)

    async def flush(self) -> None:
        """Flush write buffers, waiting until all data is written."""
        self._check_broken()
        await self._flush()

    @abstractmethod
    async def _flush(self) -> None:
        """Flush write buffers, waiting until all data is written, internal."""
        raise NotImplementedError

    async def wait_closed(self) -> None:
        """Wait until transport is fully closed."""
        await self._closed_waiter


def get_serial_classes(
    url: str,
) -> tuple[type[BaseSerial], type[BaseSerialTransport]]:
    """Get the appropriate serial and transport classes based on the URL scheme."""
    handler = get_uri_handler(url)
    return handler.sync_cls, handler.async_transport_cls


@dataclasses.dataclass
class SerialPortInfo:
    """A serial port."""

    device: str
    resolved_device: str

    vid: int | None
    pid: int | None
    serial_number: str | None
    manufacturer: str | None
    product: str | None
    bcd_device: int | None
    interface_description: str | None
    interface_num: int | None

    def __getitem__(self, key: int | slice) -> str | None:
        """Compatibility shim for `serial.tools.list_ports_common.ListPortInfo`."""
        warnings.warn(
            "Slicing `SerialPortInfo` is deprecated, use attributes instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return (str(self.device), self.description, "")[key]

    @property
    def description(self) -> str:
        """Description of the device.

        Warning:
            Deprecated, this description is unstable and will change in the future.

        """

        if self.interface_description is not None:
            product = self.product if self.product is not None else "None"
            return f"{product} - {self.interface_description}"

        if self.product is not None:
            return self.product

        return os.path.basename(self.resolved_device)


def list_serial_ports(
    platform: Platform | str = Platform.DEVICE, **kwargs: Any
) -> list[SerialPortInfo]:
    """List serial ports, defaulting to the system platform."""
    handler = get_uri_handler(platform + "://")
    return handler.list_serial_ports_func(**kwargs)


async def async_list_serial_ports(
    platform: Platform | str = Platform.DEVICE, **kwargs: Any
) -> list[SerialPortInfo]:
    """List serial ports (async), defaulting to the system platform."""
    handler = get_uri_handler(platform + "://")
    return await handler.async_list_serial_ports_func(**kwargs)


def serial_for_url(url: str, *args: Any, **kwargs: Any) -> BaseSerial:
    """Create the appropriate serial port subclass for the given URL."""
    return BaseSerial.from_url(url, *args, **kwargs)



"""Compatibility shim for pyserial's ``serial`` package backed by serialx."""

from __future__ import annotations

import importlib.util

if importlib.util.find_spec("serial.serialcli"):
    raise RuntimeError(
        "serialx-compat detected a mixed installation with pyserial files. Uninstall"
        " pyserial, pyserial-asyncio, and pyserial-asyncio-fast before installing"
        " serialx-compat."
    )

from importlib.metadata import version
from typing import Any

from serialx import (
    CR,
    EIGHTBITS,
    LF,
    PARITY_EVEN,
    PARITY_NONE,
    PARITY_ODD,
    SEVENBITS,
    STOPBITS_ONE,
    STOPBITS_TWO,
    BaseSerial,
    BaseSerialTransport,
    ModemPins,
    Parity,
    PinState,
    Serial as SerialxSerial,
    SerialException,
    SerialPortInfo,
    SerialStreamWriter,
    SerialTimeoutException,
    SerialTransport,
    StopBits,
    UnsupportedSetting,
    create_serial_connection,
    get_serial_classes,
    list_serial_ports,
    open_serial_connection,
)

VERSION = version("serialx-compat")
__version__ = VERSION

__all__ = [
    "create_serial_connection",
    "get_serial_classes",
    "list_serial_ports",
    "open_serial_connection",
    "serial_for_url",
    "ModemPins",
    "Parity",
    "PinState",
    "BaseSerial",
    "BaseSerialTransport",
    "Serial",
    "SerialException",
    "UnsupportedSetting",
    "SerialPortInfo",
    "SerialStreamWriter",
    "SerialTransport",
    "StopBits",
    "SerialTimeoutException",
    "EIGHTBITS",
    "SEVENBITS",
    "PARITY_NONE",
    "PARITY_EVEN",
    "PARITY_ODD",
    "STOPBITS_ONE",
    "STOPBITS_TWO",
    "CR",
    "LF",
    "VERSION",
    "__version__",
]


class CompatSerial(SerialxSerial):
    """Compatibility base class, maintaining runtime-compatibility with pyserial."""

    def __init__(
        self, *args: Any, _wrap_exceptions: bool = True, **kwargs: Any
    ) -> None:
        super().__init__(*args, _wrap_exceptions=_wrap_exceptions, **kwargs)

    @classmethod
    def from_url(cls, url: str, *args: Any, **kwargs: Any) -> BaseSerial:
        """Create the appropriate serial port subclass for the given URL."""
        return super().from_url(url, *args, _wrap_exceptions=True, **kwargs)


Serial = CompatSerial
serial_for_url = CompatSerial.from_url
